@@ -4,16 +4,13 @@ use crate::{
     arithmetic::util::{adc, mac, sbb},
     FieldBytes,
 };
-use core::convert::TryInto;
 use core::ops::{Add, AddAssign, Mul, MulAssign, Neg, Sub, SubAssign};
 use elliptic_curve::{
     rand_core::{CryptoRng, RngCore},
     subtle::{Choice, ConditionallySelectable, ConstantTimeEq, CtOption},
+    zeroize::DefaultIsZeroes,
 };
 
-#[cfg(feature = "zeroize")]
-use elliptic_curve::zeroize::Zeroize;
-
 /// The number of 64-bit limbs used to represent a [`FieldElement`].
 const LIMBS: usize = 4;
 
@@ -80,6 +77,8 @@ impl Default for FieldElement {
     }
 }
 
+impl DefaultIsZeroes for FieldElement {}
+
 impl FieldElement {
     /// Returns the zero element.
     pub const fn zero() -> FieldElement {
@@ -135,11 +134,11 @@ impl FieldElement {
         let is_some = (borrow as u8) & 1;
 
         // Convert w to Montgomery form: w * R^2 * R^-1 mod p = wR mod p
-        CtOption::new(FieldElement(w).mul(&R2), Choice::from(is_some))
+        CtOption::new(FieldElement(w).as_montgomery(), Choice::from(is_some))
     }
 
     /// Returns the SEC1 encoding of this field element.
-    pub fn to_bytes(&self) -> FieldBytes {
+    pub fn to_bytes(self) -> FieldBytes {
         // Convert from Montgomery form to canonical form
         let tmp =
             FieldElement::montgomery_reduce(self.0[0], self.0[1], self.0[2], self.0[3], 0, 0, 0, 0);
@@ -180,7 +179,7 @@ impl FieldElement {
         let (w3, w4) = adc(self.0[3], rhs.0[3], carry);
 
         // Attempt to subtract the modulus, to ensure the result is in the field.
-        Self::sub_inner(
+        let (result, _) = Self::sub_inner(
             w0,
             w1,
             w2,
@@ -191,7 +190,8 @@ impl FieldElement {
             MODULUS.0[2],
             MODULUS.0[3],
             0,
-        )
+        );
+        result
     }
 
     /// Returns 2*self.
@@ -201,12 +201,20 @@ impl FieldElement {
 
     /// Returns self - rhs mod p
     pub const fn subtract(&self, rhs: &Self) -> Self {
+        let (result, _) = Self::sub_inner(
+            self.0[0], self.0[1], self.0[2], self.0[3], 0, rhs.0[0], rhs.0[1], rhs.0[2], rhs.0[3],
+            0,
+        );
+        result
+    }
+
+    /// Returns self - rhs mod p
+    pub(crate) const fn informed_subtract(&self, rhs: &Self) -> (Self, u64) {
         Self::sub_inner(
             self.0[0], self.0[1], self.0[2], self.0[3], 0, rhs.0[0], rhs.0[1], rhs.0[2], rhs.0[3],
             0,
         )
     }
-
     #[inline]
     #[allow(clippy::too_many_arguments)]
     const fn sub_inner(
@@ -220,7 +228,7 @@ impl FieldElement {
         r2: u64,
         r3: u64,
         r4: u64,
-    ) -> Self {
+    ) -> (Self, u64) {
         let (w0, borrow) = sbb(l0, r0, 0);
         let (w1, borrow) = sbb(l1, r1, borrow);
         let (w2, borrow) = sbb(l2, r2, borrow);
@@ -235,7 +243,7 @@ impl FieldElement {
         let (w2, carry) = adc(w2, MODULUS.0[2] & borrow, carry);
         let (w3, _) = adc(w3, MODULUS.0[3] & borrow, carry);
 
-        FieldElement([w0, w1, w2, w3])
+        (FieldElement([w0, w1, w2, w3]), borrow)
     }
 
     /// Montgomery Reduction
@@ -310,7 +318,7 @@ impl FieldElement {
         let (r7, r8) = adc(r7, carry2, carry);
 
         // Result may be within MODULUS of the correct value
-        Self::sub_inner(
+        let (result, _) = Self::sub_inner(
             r4,
             r5,
             r6,
@@ -321,7 +329,16 @@ impl FieldElement {
             MODULUS.0[2],
             MODULUS.0[3],
             0,
-        )
+        );
+        result
+    }
+
+    pub(crate) const fn as_canonical(&self) -> Self {
+        FieldElement::montgomery_reduce(self.0[0], self.0[1], self.0[2], self.0[3], 0, 0, 0, 0)
+    }
+
+    pub(crate) const fn as_montgomery(&self) -> Self {
+        self.mul(&R2)
     }
 
     /// Returns self * rhs mod p
@@ -501,13 +518,6 @@ impl<'a> Neg for &'a FieldElement {
     }
 }
 
-#[cfg(feature = "zeroize")]
-impl Zeroize for FieldElement {
-    fn zeroize(&mut self) {
-        self.0.zeroize();
-    }
-}
-
 #[cfg(test)]
 mod tests {
     use super::FieldElement;
