@@ -7,21 +7,28 @@ use core::{
     ops::{Add, AddAssign, Mul, MulAssign, Neg, Sub, SubAssign},
 };
 use elliptic_curve::{
+    bigint::Limb,
     group::{
         ff::Field,
         prime::{PrimeCurve, PrimeCurveAffine, PrimeGroup},
         Curve, Group, GroupEncoding,
     },
+    ops::LinearCombination,
     rand_core::RngCore,
     sec1::{FromEncodedPoint, ToEncodedPoint},
     subtle::{Choice, ConditionallySelectable, ConstantTimeEq, CtOption},
-    ProjectiveArithmetic,
+    zeroize::DefaultIsZeroes,
+    PrimeCurveArithmetic, ProjectiveArithmetic,
 };
 
 impl ProjectiveArithmetic for NistP256 {
     type ProjectivePoint = ProjectivePoint;
 }
 
+impl PrimeCurveArithmetic for NistP256 {
+    type CurveGroup = ProjectivePoint;
+}
+
 /// A point on the secp256r1 curve in projective coordinates.
 #[derive(Clone, Copy, Debug)]
 #[cfg_attr(docsrs, doc(cfg(feature = "arithmetic")))]
@@ -60,7 +67,7 @@ impl GroupEncoding for ProjectivePoint {
     type Repr = CompressedPoint;
 
     fn from_bytes(bytes: &Self::Repr) -> CtOption<Self> {
-        <AffinePoint as GroupEncoding>::from_bytes(bytes).map(|point| point.into())
+        <AffinePoint as GroupEncoding>::from_bytes(bytes).map(Into::into)
     }
 
     fn from_bytes_unchecked(bytes: &Self::Repr) -> CtOption<Self> {
@@ -69,7 +76,7 @@ impl GroupEncoding for ProjectivePoint {
     }
 
     fn to_bytes(&self) -> Self::Repr {
-        CompressedPoint::clone_from_slice(self.to_affine().to_encoded_point(true).as_bytes())
+        self.to_affine().to_bytes()
     }
 }
 
@@ -87,6 +94,8 @@ impl PrimeCurve for ProjectivePoint {
     type Affine = AffinePoint;
 }
 
+impl LinearCombination for ProjectivePoint {}
+
 impl From<AffinePoint> for ProjectivePoint {
     fn from(p: AffinePoint) -> Self {
         let projective = ProjectivePoint {
@@ -105,7 +114,7 @@ impl From<ProjectivePoint> for AffinePoint {
 }
 
 impl FromEncodedPoint<NistP256> for ProjectivePoint {
-    fn from_encoded_point(p: &EncodedPoint) -> Option<Self> {
+    fn from_encoded_point(p: &EncodedPoint) -> CtOption<Self> {
         AffinePoint::from_encoded_point(p).map(ProjectivePoint::from)
     }
 }
@@ -132,14 +141,16 @@ impl ConstantTimeEq for ProjectivePoint {
     }
 }
 
+impl DefaultIsZeroes for ProjectivePoint {}
+
+impl Eq for ProjectivePoint {}
+
 impl PartialEq for ProjectivePoint {
     fn eq(&self, other: &Self) -> bool {
         self.ct_eq(other).into()
     }
 }
 
-impl Eq for ProjectivePoint {}
-
 impl ProjectivePoint {
     /// Returns the additive identity of P-256, also known as the "neutral element" or
     /// "point at infinity".
@@ -284,10 +295,10 @@ impl ProjectivePoint {
     fn mul(&self, k: &Scalar) -> ProjectivePoint {
         let mut ret = ProjectivePoint::identity();
 
-        for limb in k.0.iter().rev() {
-            for i in (0..64).rev() {
+        for limb in k.limbs().iter().rev() {
+            for i in (0..Limb::BIT_SIZE).rev() {
                 ret = ret.double();
-                ret.conditional_assign(&(ret + self), Choice::from(((limb >> i) & 1u64) as u8));
+                ret.conditional_assign(&(ret + self), Choice::from(((limb.0 >> i) & 1) as u8));
             }
         }
 
@@ -513,7 +524,7 @@ impl<'a> Neg for &'a ProjectivePoint {
 mod tests {
     use super::{AffinePoint, ProjectivePoint, Scalar};
     use crate::test_vectors::group::{ADD_TEST_VECTORS, MUL_TEST_VECTORS};
-    use elliptic_curve::group::{ff::PrimeField, prime::PrimeCurveAffine};
+    use elliptic_curve::group::{ff::PrimeField, prime::PrimeCurveAffine, GroupEncoding};
 
     #[test]
     fn affine_to_projective() {
@@ -657,4 +668,10 @@ mod tests {
             assert_eq!(res.y.to_bytes(), coords.1.into());
         }
     }
+
+    #[test]
+    fn projective_identity_to_bytes() {
+        // This is technically an invalid SEC1 encoding, but is preferable to panicking.
+        assert_eq!([0; 33], ProjectivePoint::identity().to_bytes().as_slice());
+    }
 }
