@@ -1,66 +1,41 @@
 //! Scalar field arithmetic modulo n = 115792089210356248762697446949407573529996955224135760342422259061068512044369
 
-pub mod blinding;
+pub mod blinded;
 
 use crate::{
     arithmetic::util::{adc, mac, sbb},
-    FieldBytes, NistP256,
-};
-use core::{
-    convert::TryInto,
-    ops::{Add, AddAssign, Mul, MulAssign, Neg, Sub, SubAssign},
+    FieldBytes, NistP256, SecretKey,
 };
+use core::ops::{Add, AddAssign, Mul, MulAssign, Neg, Sub, SubAssign};
 use elliptic_curve::{
+    bigint::{prelude::*, Limb, U256},
     generic_array::arr,
     group::ff::{Field, PrimeField},
+    ops::Reduce,
     rand_core::RngCore,
-    subtle::{Choice, ConditionallySelectable, ConstantTimeEq, CtOption},
-    ScalarArithmetic,
+    subtle::{
+        Choice, ConditionallySelectable, ConstantTimeEq, ConstantTimeGreater, ConstantTimeLess,
+        CtOption,
+    },
+    zeroize::DefaultIsZeroes,
+    Curve, IsHigh, ScalarArithmetic, ScalarCore,
 };
 
 #[cfg(feature = "bits")]
 use {crate::ScalarBits, elliptic_curve::group::ff::PrimeFieldBits};
 
-#[cfg(feature = "digest")]
-use ecdsa_core::{elliptic_curve::consts::U32, hazmat::FromDigest, signature::digest::Digest};
+#[cfg(feature = "serde")]
+use elliptic_curve::serde::{de, ser, Deserialize, Serialize};
 
-#[cfg(feature = "zeroize")]
-use crate::SecretKey;
-#[cfg(feature = "zeroize")]
-use elliptic_curve::zeroize::Zeroize;
-
-/// The number of 64-bit limbs used to represent a [`Scalar`].
-const LIMBS: usize = 4;
-
-type U256 = [u64; LIMBS];
+/// Array containing 4 x 64-bit unsigned integers.
+// TODO(tarcieri): replace this entirely with `U256`
+type U64x4 = [u64; 4];
 
 /// Constant representing the modulus
 /// n = FFFFFFFF 00000000 FFFFFFFF FFFFFFFF BCE6FAAD A7179E84 F3B9CAC2 FC632551
-///
-/// # Calculating the modulus
-/// One way to calculate the modulus is with `GP/PARI`:
-///
-/// ```text
-/// p = (2^224) * (2^32 - 1) + 2^192 + 2^96 - 1
-/// b = 41058363725152142129326129780047268409114441015993725554835256314039467401291
-/// E = ellinit([Mod(-3, p), Mod(b, p)])
-/// default(parisize, 120000000)
-/// n = ellsea(E)
-/// isprime(n)
-/// ```
-pub const MODULUS: U256 = [
-    0xf3b9_cac2_fc63_2551,
-    0xbce6_faad_a717_9e84,
-    0xffff_ffff_ffff_ffff,
-    0xffff_ffff_0000_0000,
-];
+const MODULUS: U64x4 = u256_to_u64x4(NistP256::ORDER);
 
-const MODULUS_SHR1: U256 = [
-    0x79dc_e561_7e31_92a8,
-    0xde73_7d56_d38b_cf42,
-    0x7fff_ffff_ffff_ffff,
-    0x7fff_ffff_8000_0000,
-];
+const FRAC_MODULUS_2: Scalar = Scalar(NistP256::ORDER.shr_vartime(1));
 
 /// MU = floor(2^512 / n)
 ///    = 115792089264276142090721624801893421302707618245269942344307673200490803338238
@@ -100,6 +75,14 @@ impl ScalarArithmetic for NistP256 {
 ///   operations over field elements represented as bits (requires `bits` feature)
 ///
 /// Please see the documentation for the relevant traits for more information.
+///
+/// # `serde` support
+///
+/// When the `serde` feature of this crate is enabled, the `Serialize` and
+/// `Deserialize` traits are impl'd for this type.
+///
+/// The serialization is a fixed-width big endian encoding. When used with
+/// textual formats, the binary data is encoded as hexadecimal.
 #[derive(Clone, Copy, Debug, Default)]
 #[cfg_attr(docsrs, doc(cfg(feature = "arithmetic")))]
 pub struct Scalar(pub(crate) U256);
@@ -119,22 +102,18 @@ impl Field for Scalar {
         // iterations is vanishingly small.
         loop {
             rng.fill_bytes(&mut bytes);
-            if let Some(scalar) = Scalar::from_repr(bytes) {
+            if let Some(scalar) = Scalar::from_repr(bytes).into() {
                 return scalar;
             }
         }
     }
 
     fn zero() -> Self {
-        Scalar::zero()
+        Self::ZERO
     }
 
     fn one() -> Self {
-        Scalar::one()
-    }
-
-    fn is_zero(&self) -> bool {
-        self.is_zero().into()
+        Self::ONE
     }
 
     #[must_use]
@@ -151,9 +130,46 @@ impl Field for Scalar {
         Scalar::invert(self)
     }
 
-    // TODO(tarcieri); stub! See: https://github.com/RustCrypto/elliptic-curves/issues/170
+    /// Tonelli-Shank's algorithm for q mod 16 = 1
+    /// <https://eprint.iacr.org/2012/685.pdf> (page 12, algorithm 5)
+    #[allow(clippy::many_single_char_names)]
     fn sqrt(&self) -> CtOption<Self> {
-        todo!("see RustCrypto/elliptic-curves#170");
+        // Note: `pow_vartime` is constant-time with respect to `self`
+        let w = self.pow_vartime(&[
+            0x279dce5617e3192a,
+            0xfde737d56d38bcf4,
+            0x07ffffffffffffff,
+            0x07fffffff8000000,
+        ]);
+
+        let mut v = Self::S;
+        let mut x = *self * w;
+        let mut b = x * w;
+        let mut z = Self::root_of_unity();
+
+        for max_v in (1..=Self::S).rev() {
+            let mut k = 1;
+            let mut tmp = b.square();
+            let mut j_less_than_v = Choice::from(1);
+
+            for j in 2..max_v {
+                let tmp_is_one = tmp.ct_eq(&Self::one());
+                let squared = Self::conditional_select(&tmp, &z, tmp_is_one).square();
+                tmp = Self::conditional_select(&squared, &tmp, tmp_is_one);
+                let new_z = Self::conditional_select(&z, &squared, tmp_is_one);
+                j_less_than_v &= !j.ct_eq(&v);
+                k = u32::conditional_select(&j, &k, tmp_is_one);
+                z = Self::conditional_select(&z, &new_z, j_less_than_v);
+            }
+
+            let result = x * z;
+            x = Self::conditional_select(&result, &x, b.ct_eq(&Self::one()));
+            z = z.square();
+            b *= z;
+            v = k;
+        }
+
+        CtOption::new(x, x.square().ct_eq(self))
     }
 }
 
@@ -168,32 +184,17 @@ impl PrimeField for Scalar {
     ///
     /// Returns None if the byte array does not contain a big-endian integer in the range
     /// [0, p).
-    fn from_repr(bytes: FieldBytes) -> Option<Self> {
-        let mut w = [0u64; LIMBS];
-
-        // Interpret the bytes as a big-endian integer w.
-        w[3] = u64::from_be_bytes(bytes[0..8].try_into().unwrap());
-        w[2] = u64::from_be_bytes(bytes[8..16].try_into().unwrap());
-        w[1] = u64::from_be_bytes(bytes[16..24].try_into().unwrap());
-        w[0] = u64::from_be_bytes(bytes[24..32].try_into().unwrap());
-
-        // If w is in the range [0, n) then w - n will overflow, resulting in a borrow
-        // value of 2^64 - 1.
-        let (_, borrow) = sbb(w[0], MODULUS[0], 0);
-        let (_, borrow) = sbb(w[1], MODULUS[1], borrow);
-        let (_, borrow) = sbb(w[2], MODULUS[2], borrow);
-        let (_, borrow) = sbb(w[3], MODULUS[3], borrow);
-        let is_some = (borrow as u8) & 1;
-
-        CtOption::new(Scalar(w), Choice::from(is_some)).into()
+    fn from_repr(bytes: FieldBytes) -> CtOption<Self> {
+        let inner = U256::from_be_byte_array(bytes);
+        CtOption::new(Self(inner), inner.ct_lt(&NistP256::ORDER))
     }
 
     fn to_repr(&self) -> FieldBytes {
         self.to_bytes()
     }
 
-    fn is_odd(&self) -> bool {
-        self.0[0] as u8 == 1
+    fn is_odd(&self) -> Choice {
+        self.0.is_odd()
     }
 
     fn multiplicative_generator() -> Self {
@@ -204,7 +205,7 @@ impl PrimeField for Scalar {
         Scalar::from_repr(arr![u8;
             0xff, 0xc9, 0x7f, 0x06, 0x2a, 0x77, 0x09, 0x92, 0xba, 0x80, 0x7a, 0xce, 0x84, 0x2a,
             0x3d, 0xfc, 0x15, 0x46, 0xca, 0xd0, 0x04, 0x37, 0x8d, 0xaf, 0x05, 0x92, 0xd7, 0xfb,
-            0xb4, 0x1e, 0x66, 0x00,
+            0xb4, 0x1e, 0x66, 0x02,
         ])
         .unwrap()
     }
@@ -223,156 +224,26 @@ impl PrimeFieldBits for Scalar {
         self.into()
     }
 
-    #[cfg(target_pointer_width = "32")]
     fn char_le_bits() -> ScalarBits {
-        [
-            0xfc63_2551,
-            0xf3b9_cac2,
-            0xa717_9e84,
-            0xbce6_faad,
-            0xffff_ffff,
-            0xffff_ffff,
-            0x0000_0000,
-            0xffff_ffff,
-        ]
-        .into()
-    }
-
-    #[cfg(target_pointer_width = "64")]
-    fn char_le_bits() -> ScalarBits {
-        MODULUS.into()
-    }
-}
-
-impl From<u64> for Scalar {
-    fn from(k: u64) -> Self {
-        Scalar([k, 0, 0, 0])
-    }
-}
-
-impl PartialEq for Scalar {
-    fn eq(&self, other: &Self) -> bool {
-        self.ct_eq(other).into()
-    }
-}
-
-impl Eq for Scalar {}
-
-impl PartialOrd for Scalar {
-    fn partial_cmp(&self, other: &Self) -> Option<core::cmp::Ordering> {
-        Some(self.cmp(other))
-    }
-}
-
-/// Returns sign of left - right
-fn cmp_vartime(left: &U256, right: &U256) -> i32 {
-    use core::cmp::Ordering::*;
-
-    // since we're little-endian, need to reverse iterate
-    for (l, r) in left.iter().rev().zip(right.iter().rev()) {
-        match l.cmp(r) {
-            Less => return -1,
-            Greater => return 1,
-            Equal => continue,
-        }
-    }
-    0
-}
-
-fn shr1(u256: &mut U256) {
-    let mut bit: u64 = 0;
-    for digit in u256.iter_mut().rev() {
-        let new_digit = (bit << 63) | (*digit >> 1);
-        bit = *digit & 1;
-        *digit = new_digit;
-    }
-}
-
-impl Ord for Scalar {
-    fn cmp(&self, other: &Self) -> core::cmp::Ordering {
-        use core::cmp::Ordering::*;
-        match cmp_vartime(&self.0, &other.0) {
-            -1 => Less,
-            0 => Equal,
-            1 => Greater,
-            _ => unreachable!(),
-        }
-    }
-}
-
-#[cfg(feature = "digest")]
-#[cfg_attr(docsrs, doc(cfg(feature = "digest")))]
-impl FromDigest<NistP256> for Scalar {
-    /// Convert the output of a digest algorithm into a [`Scalar`] reduced
-    /// modulo n.
-    fn from_digest<D>(digest: D) -> Self
-    where
-        D: Digest<OutputSize = U32>,
-    {
-        Self::from_bytes_reduced(&digest.finalize())
+        NistP256::ORDER.to_uint_array().into()
     }
 }
 
 impl Scalar {
-    /// Returns the zero scalar.
-    pub const fn zero() -> Scalar {
-        Scalar([0, 0, 0, 0])
-    }
+    /// Zero scalar.
+    pub const ZERO: Self = Self(U256::ZERO);
 
-    /// Returns the multiplicative identity.
-    pub const fn one() -> Scalar {
-        Scalar([1, 0, 0, 0])
-    }
-
-    /// Parses the given byte array as a scalar.
-    ///
-    /// Subtracts the modulus when the byte array is larger than the modulus.
-    pub fn from_bytes_reduced(bytes: &FieldBytes) -> Self {
-        Self::sub_inner(
-            u64::from_be_bytes(bytes[24..32].try_into().unwrap()),
-            u64::from_be_bytes(bytes[16..24].try_into().unwrap()),
-            u64::from_be_bytes(bytes[8..16].try_into().unwrap()),
-            u64::from_be_bytes(bytes[0..8].try_into().unwrap()),
-            0,
-            MODULUS[0],
-            MODULUS[1],
-            MODULUS[2],
-            MODULUS[3],
-            0,
-        )
-    }
+    /// Multiplicative identity.
+    pub const ONE: Self = Self(U256::ONE);
 
     /// Returns the SEC1 encoding of this scalar.
     pub fn to_bytes(&self) -> FieldBytes {
-        let mut ret = FieldBytes::default();
-        ret[0..8].copy_from_slice(&self.0[3].to_be_bytes());
-        ret[8..16].copy_from_slice(&self.0[2].to_be_bytes());
-        ret[16..24].copy_from_slice(&self.0[1].to_be_bytes());
-        ret[24..32].copy_from_slice(&self.0[0].to_be_bytes());
-        ret
-    }
-
-    /// Determine if this `Scalar` is zero.
-    ///
-    /// # Returns
-    ///
-    /// If zero, return `Choice(1)`.  Otherwise, return `Choice(0)`.
-    pub fn is_zero(&self) -> Choice {
-        self.ct_eq(&Scalar::zero())
+        self.0.to_be_byte_array()
     }
 
     /// Returns self + rhs mod n
     pub const fn add(&self, rhs: &Self) -> Self {
-        // Bit 256 of n is set, so addition can result in five words.
-        let (w0, carry) = adc(self.0[0], rhs.0[0], 0);
-        let (w1, carry) = adc(self.0[1], rhs.0[1], carry);
-        let (w2, carry) = adc(self.0[2], rhs.0[2], carry);
-        let (w3, w4) = adc(self.0[3], rhs.0[3], carry);
-
-        // Attempt to subtract the modulus, to ensure the result is in the field.
-        Self::sub_inner(
-            w0, w1, w2, w3, w4, MODULUS[0], MODULUS[1], MODULUS[2], MODULUS[3], 0,
-        )
+        Self(self.0.add_mod(&rhs.0, &NistP256::ORDER))
     }
 
     /// Returns 2*self.
@@ -380,43 +251,109 @@ impl Scalar {
         self.add(self)
     }
 
-    /// Returns self - rhs mod n
-    pub const fn subtract(&self, rhs: &Self) -> Self {
-        Self::sub_inner(
-            self.0[0], self.0[1], self.0[2], self.0[3], 0, rhs.0[0], rhs.0[1], rhs.0[2], rhs.0[3],
-            0,
-        )
+    /// Returns self - rhs mod n.
+    pub const fn sub(&self, rhs: &Self) -> Self {
+        Self(self.0.sub_mod(&rhs.0, &NistP256::ORDER))
     }
 
-    #[inline]
-    #[allow(clippy::too_many_arguments)]
-    const fn sub_inner(
-        l0: u64,
-        l1: u64,
-        l2: u64,
-        l3: u64,
-        l4: u64,
-        r0: u64,
-        r1: u64,
-        r2: u64,
-        r3: u64,
-        r4: u64,
-    ) -> Self {
-        let (w0, borrow) = sbb(l0, r0, 0);
-        let (w1, borrow) = sbb(l1, r1, borrow);
-        let (w2, borrow) = sbb(l2, r2, borrow);
-        let (w3, borrow) = sbb(l3, r3, borrow);
-        let (_, borrow) = sbb(l4, r4, borrow);
+    /// Returns self * rhs mod n
+    pub const fn mul(&self, rhs: &Self) -> Self {
+        let (lo, hi) = self.0.mul_wide(&rhs.0);
+        Self::barrett_reduce(lo, hi)
+    }
 
-        // If underflow occurred on the final limb, borrow = 0xfff...fff, otherwise
-        // borrow = 0x000...000. Thus, we use it as a mask to conditionally add the
-        // modulus.
-        let (w0, carry) = adc(w0, MODULUS[0] & borrow, 0);
-        let (w1, carry) = adc(w1, MODULUS[1] & borrow, carry);
-        let (w2, carry) = adc(w2, MODULUS[2] & borrow, carry);
-        let (w3, _) = adc(w3, MODULUS[3] & borrow, carry);
+    /// Returns self * self mod p
+    pub const fn square(&self) -> Self {
+        // Schoolbook multiplication.
+        self.mul(self)
+    }
 
-        Scalar([w0, w1, w2, w3])
+    /// Returns the multiplicative inverse of self, if self is non-zero
+    pub fn invert(&self) -> CtOption<Self> {
+        // We need to find b such that b * a ≡ 1 mod p. As we are in a prime
+        // field, we can apply Fermat's Little Theorem:
+        //
+        //    a^p         ≡ a mod p
+        //    a^(p-1)     ≡ 1 mod p
+        //    a^(p-2) * a ≡ 1 mod p
+        //
+        // Thus inversion can be implemented with a single exponentiation.
+        //
+        // This is `n - 2`, so the top right two digits are `4f` instead of `51`.
+        let inverse = self.pow_vartime(&[
+            0xf3b9_cac2_fc63_254f,
+            0xbce6_faad_a717_9e84,
+            0xffff_ffff_ffff_ffff,
+            0xffff_ffff_0000_0000,
+        ]);
+
+        CtOption::new(inverse, !self.is_zero())
+    }
+
+    /// Faster inversion using Stein's algorithm
+    #[allow(non_snake_case)]
+    pub fn invert_vartime(&self) -> CtOption<Self> {
+        // https://link.springer.com/article/10.1007/s13389-016-0135-4
+
+        let mut u = *self;
+        // currently an invalid scalar
+        let mut v = Scalar(NistP256::ORDER);
+        let mut A = Self::one();
+        let mut C = Self::zero();
+
+        while !bool::from(u.is_zero()) {
+            // u-loop
+            while bool::from(u.is_even()) {
+                u.shr1();
+
+                let was_odd: bool = A.is_odd().into();
+                A.shr1();
+
+                if was_odd {
+                    A += FRAC_MODULUS_2;
+                    A += Self::one();
+                }
+            }
+
+            // v-loop
+            while bool::from(v.is_even()) {
+                v.shr1();
+
+                let was_odd: bool = C.is_odd().into();
+                C.shr1();
+
+                if was_odd {
+                    C += FRAC_MODULUS_2;
+                    C += Self::one();
+                }
+            }
+
+            // sub-step
+            if u >= v {
+                u -= &v;
+                A -= &C;
+            } else {
+                v -= &u;
+                C -= &A;
+            }
+        }
+
+        CtOption::new(C, !self.is_zero())
+    }
+
+    /// Is integer representing equivalence class odd?
+    pub fn is_odd(&self) -> Choice {
+        self.0.is_odd()
+    }
+
+    /// Is integer representing equivalence class even?
+    pub fn is_even(&self) -> Choice {
+        !self.is_odd()
+    }
+
+    /// Borrow the inner limbs array.
+    pub(crate) const fn limbs(&self) -> &[Limb] {
+        self.0.limbs()
     }
 
     /// Barrett Reduction
@@ -449,16 +386,17 @@ impl Scalar {
     ///   https://csrc.nist.gov/csrc/media/events/workshop-on-elliptic-curve-cryptography-standards/documents/papers/session6-adalier-mehmet.pdf
     #[inline]
     #[allow(clippy::too_many_arguments)]
-    const fn barrett_reduce(
-        a0: u64,
-        a1: u64,
-        a2: u64,
-        a3: u64,
-        a4: u64,
-        a5: u64,
-        a6: u64,
-        a7: u64,
-    ) -> Self {
+    const fn barrett_reduce(lo: U256, hi: U256) -> Self {
+        let lo = u256_to_u64x4(lo);
+        let hi = u256_to_u64x4(hi);
+        let a0 = lo[0];
+        let a1 = lo[1];
+        let a2 = lo[2];
+        let a3 = lo[3];
+        let a4 = hi[0];
+        let a5 = hi[1];
+        let a6 = hi[2];
+        let a7 = hi[3];
         let q1: [u64; 5] = [a3, a4, a5, a6, a7];
 
         const fn q1_times_mu_shift_five(q1: &[u64; 5]) -> [u64; 5] {
@@ -570,145 +508,103 @@ impl Scalar {
         // and 90% of the time, no subtraction will be needed.
         let r = subtract_n_if_necessary(r[0], r[1], r[2], r[3], r[4]);
         let r = subtract_n_if_necessary(r[0], r[1], r[2], r[3], r[4]);
-        Scalar([r[0], r[1], r[2], r[3]])
+        Scalar::from_u64x4_unchecked([r[0], r[1], r[2], r[3]])
     }
 
-    /// Returns self * rhs mod n
-    pub const fn mul(&self, rhs: &Self) -> Self {
-        // Schoolbook multiplication.
-
-        let (w0, carry) = mac(0, self.0[0], rhs.0[0], 0);
-        let (w1, carry) = mac(0, self.0[0], rhs.0[1], carry);
-        let (w2, carry) = mac(0, self.0[0], rhs.0[2], carry);
-        let (w3, w4) = mac(0, self.0[0], rhs.0[3], carry);
-
-        let (w1, carry) = mac(w1, self.0[1], rhs.0[0], 0);
-        let (w2, carry) = mac(w2, self.0[1], rhs.0[1], carry);
-        let (w3, carry) = mac(w3, self.0[1], rhs.0[2], carry);
-        let (w4, w5) = mac(w4, self.0[1], rhs.0[3], carry);
-
-        let (w2, carry) = mac(w2, self.0[2], rhs.0[0], 0);
-        let (w3, carry) = mac(w3, self.0[2], rhs.0[1], carry);
-        let (w4, carry) = mac(w4, self.0[2], rhs.0[2], carry);
-        let (w5, w6) = mac(w5, self.0[2], rhs.0[3], carry);
-
-        let (w3, carry) = mac(w3, self.0[3], rhs.0[0], 0);
-        let (w4, carry) = mac(w4, self.0[3], rhs.0[1], carry);
-        let (w5, carry) = mac(w5, self.0[3], rhs.0[2], carry);
-        let (w6, w7) = mac(w6, self.0[3], rhs.0[3], carry);
-
-        Scalar::barrett_reduce(w0, w1, w2, w3, w4, w5, w6, w7)
-    }
-
-    /// Returns self * self mod p
-    pub const fn square(&self) -> Self {
-        // Schoolbook multiplication.
-        self.mul(self)
-    }
-
-    /// Returns `self^by`, where `by` is a little-endian integer exponent.
+    /// Perform unchecked conversion from a U64x4 to a Scalar.
     ///
-    /// **This operation is variable time with respect to the exponent.** If the exponent
-    /// is fixed, this operation is effectively constant time.
-    pub fn pow_vartime(&self, by: &[u64; 4]) -> Self {
-        let mut res = Self::one();
-        for e in by.iter().rev() {
-            for i in (0..64).rev() {
-                res = res.square();
-
-                if ((*e >> i) & 1) == 1 {
-                    res *= self;
-                }
-            }
-        }
-        res
+    /// Note: this does *NOT* ensure that the provided value is less than `MODULUS`.
+    // TODO(tarcieri): implement all algorithms in terms of `U256`?
+    #[cfg(target_pointer_width = "32")]
+    const fn from_u64x4_unchecked(limbs: U64x4) -> Self {
+        Self(U256::from_uint_array([
+            (limbs[0] & 0xFFFFFFFF) as u32,
+            (limbs[0] >> 32) as u32,
+            (limbs[1] & 0xFFFFFFFF) as u32,
+            (limbs[1] >> 32) as u32,
+            (limbs[2] & 0xFFFFFFFF) as u32,
+            (limbs[2] >> 32) as u32,
+            (limbs[3] & 0xFFFFFFFF) as u32,
+            (limbs[3] >> 32) as u32,
+        ]))
     }
 
-    /// Returns the multiplicative inverse of self, if self is non-zero
-    pub fn invert(&self) -> CtOption<Self> {
-        // We need to find b such that b * a ≡ 1 mod p. As we are in a prime
-        // field, we can apply Fermat's Little Theorem:
-        //
-        //    a^p         ≡ a mod p
-        //    a^(p-1)     ≡ 1 mod p
-        //    a^(p-2) * a ≡ 1 mod p
-        //
-        // Thus inversion can be implemented with a single exponentiation.
-        //
-        // This is `n - 2`, so the top right two digits are `4f` instead of `51`.
-        let inverse = self.pow_vartime(&[
-            0xf3b9_cac2_fc63_254f,
-            0xbce6_faad_a717_9e84,
-            0xffff_ffff_ffff_ffff,
-            0xffff_ffff_0000_0000,
-        ]);
-
-        CtOption::new(inverse, !self.is_zero())
-    }
-
-    /// Is integer representing equivalence class odd
-    pub fn is_odd(&self) -> Choice {
-        ((self.0[0] & 1) as u8).into()
-    }
-
-    /// Is integer representing equivalence class even
-    pub fn is_even(&self) -> Choice {
-        !self.is_odd()
+    /// Perform unchecked conversion from a U64x4 to a Scalar.
+    ///
+    /// Note: this does *NOT* ensure that the provided value is less than `MODULUS`.
+    // TODO(tarcieri): implement all algorithms in terms of `U256`?
+    #[cfg(target_pointer_width = "64")]
+    const fn from_u64x4_unchecked(limbs: U64x4) -> Self {
+        Self(U256::from_uint_array(limbs))
     }
 
+    /// Shift right by one bit
     fn shr1(&mut self) {
-        shr1(&mut self.0);
+        self.0 >>= 1;
     }
+}
 
-    /// Faster inversion using Stein's algorithm
-    pub fn invert_vartime(&self) -> CtOption<Self> {
-        // https://link.springer.com/article/10.1007/s13389-016-0135-4
+impl DefaultIsZeroes for Scalar {}
 
-        let mut u = *self;
-        // currently an invalid scalar
-        let mut v = Scalar(MODULUS);
+impl Eq for Scalar {}
 
-        #[allow(non_snake_case)]
-        let mut A = Self::one();
-        #[allow(non_snake_case)]
-        let mut C = Self::zero();
+impl IsHigh for Scalar {
+    fn is_high(&self) -> Choice {
+        self.0.ct_gt(&FRAC_MODULUS_2.0)
+    }
+}
 
-        while !bool::from(u.is_zero()) {
-            // u-loop
-            while bool::from(u.is_even()) {
-                u.shr1();
-                if bool::from(A.is_even()) {
-                    A.shr1();
-                } else {
-                    A.shr1();
-                    A += Scalar(MODULUS_SHR1);
-                    A += Self::one();
-                }
-            }
+impl PartialEq for Scalar {
+    fn eq(&self, other: &Self) -> bool {
+        self.ct_eq(other).into()
+    }
+}
 
-            // v-loop
-            while bool::from(v.is_even()) {
-                v.shr1();
-                if bool::from(C.is_even()) {
-                    C.shr1();
-                } else {
-                    C.shr1();
-                    C += Scalar(MODULUS_SHR1);
-                    C += Self::one();
-                }
-            }
+impl PartialOrd for Scalar {
+    fn partial_cmp(&self, other: &Self) -> Option<core::cmp::Ordering> {
+        Some(self.cmp(other))
+    }
+}
 
-            // sub-step
-            if u >= v {
-                u -= &v;
-                A -= &C;
-            } else {
-                v -= &u;
-                C -= &A;
-            }
-        }
+impl Ord for Scalar {
+    fn cmp(&self, other: &Self) -> core::cmp::Ordering {
+        self.0.cmp(&other.0)
+    }
+}
 
-        CtOption::new(C, !self.is_zero())
+impl From<u64> for Scalar {
+    fn from(k: u64) -> Self {
+        Scalar(k.into())
+    }
+}
+
+impl From<ScalarCore<NistP256>> for Scalar {
+    fn from(scalar: ScalarCore<NistP256>) -> Scalar {
+        Scalar(*scalar.as_uint())
+    }
+}
+
+impl From<&ScalarCore<NistP256>> for Scalar {
+    fn from(scalar: &ScalarCore<NistP256>) -> Scalar {
+        Scalar(*scalar.as_uint())
+    }
+}
+
+impl From<Scalar> for ScalarCore<NistP256> {
+    fn from(scalar: Scalar) -> ScalarCore<NistP256> {
+        ScalarCore::from(&scalar)
+    }
+}
+
+impl From<&Scalar> for ScalarCore<NistP256> {
+    fn from(scalar: &Scalar) -> ScalarCore<NistP256> {
+        ScalarCore::new(scalar.0).unwrap()
+    }
+}
+
+impl From<Scalar> for U256 {
+    fn from(scalar: Scalar) -> U256 {
+        scalar.0
     }
 }
 
@@ -752,7 +648,7 @@ impl Sub<Scalar> for Scalar {
     type Output = Scalar;
 
     fn sub(self, other: Scalar) -> Scalar {
-        Scalar::subtract(&self, &other)
+        Scalar::sub(&self, &other)
     }
 }
 
@@ -760,7 +656,7 @@ impl Sub<&Scalar> for &Scalar {
     type Output = Scalar;
 
     fn sub(self, other: &Scalar) -> Scalar {
-        Scalar::subtract(self, other)
+        Scalar::sub(self, other)
     }
 }
 
@@ -768,19 +664,19 @@ impl Sub<&Scalar> for Scalar {
     type Output = Scalar;
 
     fn sub(self, other: &Scalar) -> Scalar {
-        Scalar::subtract(&self, other)
+        Scalar::sub(&self, other)
     }
 }
 
 impl SubAssign<Scalar> for Scalar {
     fn sub_assign(&mut self, rhs: Scalar) {
-        *self = Scalar::subtract(self, &rhs);
+        *self = Scalar::sub(self, &rhs);
     }
 }
 
 impl SubAssign<&Scalar> for Scalar {
     fn sub_assign(&mut self, rhs: &Scalar) {
-        *self = Scalar::subtract(self, rhs);
+        *self = Scalar::sub(self, rhs);
     }
 }
 
@@ -836,46 +732,31 @@ impl<'a> Neg for &'a Scalar {
     }
 }
 
+impl Reduce<U256> for Scalar {
+    fn from_uint_reduced(w: U256) -> Self {
+        let (r, underflow) = w.sbb(&NistP256::ORDER, Limb::ZERO);
+        let underflow = Choice::from((underflow.0 >> (Limb::BIT_SIZE - 1)) as u8);
+        Self(U256::conditional_select(&w, &r, !underflow))
+    }
+}
+
 impl ConditionallySelectable for Scalar {
     fn conditional_select(a: &Self, b: &Self, choice: Choice) -> Self {
-        Scalar([
-            u64::conditional_select(&a.0[0], &b.0[0], choice),
-            u64::conditional_select(&a.0[1], &b.0[1], choice),
-            u64::conditional_select(&a.0[2], &b.0[2], choice),
-            u64::conditional_select(&a.0[3], &b.0[3], choice),
-        ])
+        Self(U256::conditional_select(&a.0, &b.0, choice))
     }
 }
 
 impl ConstantTimeEq for Scalar {
     fn ct_eq(&self, other: &Self) -> Choice {
-        self.0[0].ct_eq(&other.0[0])
-            & self.0[1].ct_eq(&other.0[1])
-            & self.0[2].ct_eq(&other.0[2])
-            & self.0[3].ct_eq(&other.0[3])
+        self.0.ct_eq(&other.0)
     }
 }
 
-#[cfg(all(feature = "bits", target_pointer_width = "32"))]
+#[cfg(feature = "bits")]
 #[cfg_attr(docsrs, doc(cfg(feature = "bits")))]
 impl From<&Scalar> for ScalarBits {
     fn from(scalar: &Scalar) -> ScalarBits {
-        let mut output = [0u32; 8];
-
-        for (input, output) in scalar.0.iter().zip(output.chunks_mut(2)) {
-            output[0] = (input & 0xFFFFFFFF) as u32;
-            output[1] = (input >> 32) as u32;
-        }
-
-        output.into()
-    }
-}
-
-#[cfg(all(feature = "bits", target_pointer_width = "64"))]
-#[cfg_attr(docsrs, doc(cfg(feature = "bits")))]
-impl From<&Scalar> for ScalarBits {
-    fn from(scalar: &Scalar) -> ScalarBits {
-        scalar.0.into()
+        scalar.0.to_uint_array().into()
     }
 }
 
@@ -891,28 +772,60 @@ impl From<&Scalar> for FieldBytes {
     }
 }
 
-#[cfg(feature = "zeroize")]
 impl From<&SecretKey> for Scalar {
     fn from(secret_key: &SecretKey) -> Scalar {
-        *secret_key.to_secret_scalar()
+        *secret_key.to_nonzero_scalar()
     }
 }
 
-#[cfg(feature = "zeroize")]
-impl Zeroize for Scalar {
-    fn zeroize(&mut self) {
-        self.0.as_mut().zeroize()
+#[cfg(feature = "serde")]
+#[cfg_attr(docsrs, doc(cfg(feature = "serde")))]
+impl Serialize for Scalar {
+    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
+    where
+        S: ser::Serializer,
+    {
+        ScalarCore::from(self).serialize(serializer)
     }
 }
 
+#[cfg(feature = "serde")]
+#[cfg_attr(docsrs, doc(cfg(feature = "serde")))]
+impl<'de> Deserialize<'de> for Scalar {
+    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
+    where
+        D: de::Deserializer<'de>,
+    {
+        Ok(ScalarCore::deserialize(deserializer)?.into())
+    }
+}
+
+/// Convert to a [`U64x4`] array.
+// TODO(tarcieri): implement all algorithms in terms of `U256`?
+#[cfg(target_pointer_width = "32")]
+pub(crate) const fn u256_to_u64x4(u256: U256) -> U64x4 {
+    let limbs = u256.to_uint_array();
+
+    [
+        (limbs[0] as u64) | ((limbs[1] as u64) << 32),
+        (limbs[2] as u64) | ((limbs[3] as u64) << 32),
+        (limbs[4] as u64) | ((limbs[5] as u64) << 32),
+        (limbs[6] as u64) | ((limbs[7] as u64) << 32),
+    ]
+}
+
+/// Convert to a [`U64x4`] array.
+// TODO(tarcieri): implement all algorithms in terms of `U256`?
+#[cfg(target_pointer_width = "64")]
+pub(crate) const fn u256_to_u64x4(u256: U256) -> U64x4 {
+    u256.to_uint_array()
+}
+
 #[cfg(test)]
 mod tests {
     use super::Scalar;
-    use crate::FieldBytes;
-    use elliptic_curve::group::ff::PrimeField;
-
-    #[cfg(feature = "zeroize")]
-    use crate::SecretKey;
+    use crate::{FieldBytes, SecretKey};
+    use elliptic_curve::group::ff::{Field, PrimeField};
 
     #[test]
     fn from_to_bytes_roundtrip() {
@@ -924,8 +837,8 @@ mod tests {
         assert_eq!(bytes, scalar.to_bytes());
     }
 
+    /// Basic tests that multiplication works.
     #[test]
-    // Basic tests that multiplication works.
     fn multiply() {
         let one = Scalar::one();
         let two = one + &one;
@@ -941,8 +854,8 @@ mod tests {
         assert_eq!(six, minus_two * &minus_three);
     }
 
+    /// Basic tests that scalar inversion works.
     #[test]
-    // Basic tests that scalar inversion works.
     fn invert() {
         let one = Scalar::one();
         let three = one + &one + &one;
@@ -958,12 +871,21 @@ mod tests {
         assert_eq!(three * &inv_minus_three, -one);
     }
 
-    // Tests that a Scalar can be safely converted to a SecretKey and back
+    /// Basic tests that sqrt works.
+    #[test]
+    fn sqrt() {
+        for &n in &[1u64, 4, 9, 16, 25, 36, 49, 64] {
+            let scalar = Scalar::from(n);
+            let sqrt = scalar.sqrt().unwrap();
+            assert_eq!(sqrt.square(), scalar);
+        }
+    }
+
+    /// Tests that a Scalar can be safely converted to a SecretKey and back
     #[test]
-    #[cfg(feature = "zeroize")]
     fn from_ec_secret() {
         let scalar = Scalar::one();
-        let secret = SecretKey::from_bytes(scalar.to_bytes()).unwrap();
+        let secret = SecretKey::from_be_bytes(&scalar.to_bytes()).unwrap();
         let rederived_scalar = Scalar::from(&secret);
         assert_eq!(scalar.0, rederived_scalar.0);
     }
